[![CircleCI](https://circleci.com/gh/effective-spa/server_component_rails.svg?style=svg)](https://circleci.com/gh/effective-spa/server_component_rails) [![codecov](https://codecov.io/gh/effective-spa/server_component_rails/branch/master/graph/badge.svg)](https://codecov.io/gh/effective-spa/server_component_rails) [![Maintainability](https://api.codeclimate.com/v1/badges/3a8af74f091ff7c3dbf4/maintainability)](https://codeclimate.com/github/effective-spa/server_component_rails/maintainability) [![Gem Version](https://badge.fury.io/rb/server_component.svg)](https://badge.fury.io/rb/server_component)


# ServerComponent

**ServerComponent** is a painless framework for React apps.

*This gem is designed to be used with **npm package version of [server_component](https://github.com/effective-spa/server_component)**.*

### Why?

SPAs do not always require much-highly scalable technology stack. There need alternatives for those who want to keep less code with retaining expressiveness of SPA. ServerComponent is a smart one of them.

The following image depicts how it reduces the complexity among modules. You can get more development efficiency with keeping advantages of SPA as you choose ServerComponent.

![figure](https://github.com/effective-spa/server_component_rails/raw/master/figure.png)

### Installation

Add this line to your application's Gemfile and run `bundle install`:

```ruby
gem 'server_component'
```

Run this command to initialize configuration file and base classes.

```sh
$ rails generate server_component:install
```

## 🚀 Getting Started

Add a new component controller:

```rb
class Api::CounterComponentController < ServerComponent::Base
  def self.initial_state
    {
      count: 0
    }
  end

  action :increment
  def increment
    set_state do |s|
      s.count { |prev| prev + 1 }
    end
  end

  action :decrement
  def decrement
    set_state do |s|
      s.count { |prev| prev - 1 }
    end
  end
end
```

Add new lines into routes.rb:

```rb
component_scope :api do
  component :counter, actions: [:increment, :decrement]
end
```

Install [server_component](https://github.com/effective-spa/server_component) of NPM package in your frontend module.

Here is a sample client code (with Babel >= 7 and decorator):

```js
import React, { Component } from 'react';
import ReactDOM from 'react-dom';
import ServerComponent, { server_component, consumer } from 'server_component';

@server_component('counter')
class CounterContainer extends Component { }

@consumer('counter')
class CounterBody extends Component {
  render() {
    const { increment, decrement, state: { count } } = this.props.counter;
    return (
      <div>
        <h1>Counter</h1>
        <button onClick={() => increment()}>increment</button>
        <button onClick={() => decrement()}>decrement</button>
        <hr />
        <div>Count: {count}</div>
      </div>
    );
  }
}

class Main extends Component {
  render() {
    return (
      <ServerComponent.Use name="api">
        <CounterContainer>
          <CounterBody />
        </CounterContainer>
      </ServerComponent.Use>
    );
  }
}

ReactDOM.render(<Main />, document.getElementById('root'));
```

(Note that the conventional Rails' meta tags for CSRF token are needed.)

If you don't use decorators, here's an alternative:

```js
import ServerComponent, { connectServer, consume } from 'server_component';

const CounterContainer = connectServer(class extends Component {}, 'counter');

const CounterBody = consume(class extends Component {
  :
}, 'counter');
```


## 🎓 Usage

### Write view files with [Jsrb](https://github.com/effective-spa/jsrb)

[Jsrb](https://github.com/effective-spa/jsrb), a view handler that generates JavaScript, is included in this library. Because the action of the server component merely renders executable JavaScript code instead of JSON, you have to build code safely. But don't be serious! Most of state updating functions won't be much complicated. Using `Jsrb`, you can easily and safely construct JavaScript code.

#### Example

In `controllers/books_component_controller.rb`:

```rb
action :create
def create
  @book = Book.new(book_params)
  @book.save
  render :js
end
```

In `views/books_component/create.js.jsrb`:

```rb
# Note that set_state must take `js` as an argument.
set_state(js) do |s|
  s.total { |prev| prev + 1 }
  s.books do |books|
    books.concat id: @book.id, title: @book.title
  end
end
```

### Fallback to conventional JSON handling

The component controller is just a Rails controller, so you can render JSON in a conventional manner. The client side handles response by Content-Type and status properly, so you don't have to reconstruct the status handling logic.

#### Example

In `controllers/books_component_controller.rb`:

```rb
action :create
def create
  @book = Book.new(book_params)
  if @book.save
    render json: { book: { id: @book.id, title: @book.title }
  else
    render json: { errors: @book.error_messages }, status: :unprocessable_entity
  end
end
```

In frontend:

```js
@server_component('book')
class BookContainer extends Component {
  // method name sufixed with `Ok` automatically
  // executed after 2XX response
  createOk({ book }) {
    this.setState({
      books: this.state.books.concat(book),
    });
  }

  // method name sufixed with `Err` automatically
  // executed after 4XX response
  createErr({ errors }) {
    this.setState({ errors });
  }
}
```

### Utility for temporary UI state

It is a common UI pattern to display the user operated resource in a particular state while waiting the server response. In such cases, you have to `setState` in a certain component before request. ServerComponent's controller has an utility to implement this behavior.

#### Example

```rb
action :create do |c|
  c.before do |s, data|
    # data is a request payload
    s.books do |book|
      book.concat title: data[:title], creating: true
    end
  end
end
def create
  @book = Book.new(book_params)
  @book.save
  set_state do |s|
    s.books do |books|
      books = books.filter { |b| !b[:creating] }
      if @book.persisted?
        books.push id: @book.id, title: @book.title
      end
      books
    end
  end
end
```

### Uploading files

File uploading requires FormData request because its Content-Type must be `multipart/form-data`. You have to specify explicitly that the action should accept file if you need.

#### Example

```rb
action :upload do |c|
  c.accept_file!
end
def upload
  params[:file] # => ActionDispatch::Http::UploadedFile
end
```

## 🤝 Contributing

**Contributions are extremely welcome on Github Issue or Pull Request!**

## 📝 License

The gem is available as open source under the terms of the [MIT License](http://opensource.org/licenses/MIT).
